@@ -1,6 +1,7 @@
-/* diritor.h: Iterator through entries in a directory.
- *
- * Copyright (C) 2007,2008,2010,2011,2012,2013,2014,2015 Olly Betts
+/** @file diritor.h
+ * @brief Iterator through entries in a directory.
+ */
+/* Copyright (C) 2007,2008,2010,2011,2012,2013,2014,2015,2018 Olly Betts
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,10 +21,10 @@
 #ifndef OMEGA_INCLUDED_DIRITOR_H
 #define OMEGA_INCLUDED_DIRITOR_H
 
+#include <cerrno>
 #include <string>
 
 #include "safedirent.h"
-#include "safeerrno.h"
 #include "safefcntl.h"
 #include "safesysstat.h"
 #include "safeunistd.h"
@@ -41,6 +42,7 @@
 #include "common/noreturn.h"
 
 #include "loadfile.h"
+#include "md5wrap.h"
 #include "runfilter.h" // For class ReadError.
 
 struct FileNotFound { };
@@ -85,6 +87,10 @@ class DirectoryIterator {
 
     void build_path();
 
+    void open_fd();
+
+    void close_fd();
+
   public:
 
     explicit DirectoryIterator(bool follow_symlinks_)
@@ -92,7 +98,7 @@ class DirectoryIterator {
 
     ~DirectoryIterator() {
 	if (dir) closedir(dir);
-	if (fd >= 0) close(fd);
+	if (fd >= 0) close_fd();
     }
 
     /// Start iterating through entries in @a path.
@@ -106,10 +112,7 @@ class DirectoryIterator {
     //
     //  @return false if there are no more entries.
     bool next() {
-	if (fd >= 0) {
-	    close(fd);
-	    fd = -1;
-	}
+	if (fd >= 0) close_fd();
 	path.resize(path_len);
 	errno = 0;
 	do {
@@ -233,28 +236,23 @@ class DirectoryIterator {
     std::string get_magic_mimetype();
 
     std::string file_to_string() {
-	build_path();
 	std::string out;
-	int flags = NOCACHE;
-	if (try_noatime()) flags |= NOATIME;
-	fd = load_file_fd(path, out, flags);
-	if (fd < 0) {
-	    if (errno == ENOENT || errno == ENOTDIR) throw FileNotFound();
-	    throw ReadError("load_file failed");
+	if (!load_file_from_fd(get_fd(), out)) {
+	    throw ReadError("loading file failed");
 	}
 	return out;
     }
 
     std::string gzfile_to_string() {
-	build_path();
-	std::string out;
-	gzFile zfh = gzopen(path.c_str(), "rb");
-	if (zfh == NULL) {
-	    if (errno == ENOENT || errno == ENOTDIR) {
-		throw FileNotFound();
-	    }
-	    throw ReadError("gzopen() failed");
+	int dup_fd = dup(get_fd());
+	if (fd < 0) {
+	    throw ReadError("dup() failed");
 	}
+	gzFile zfh = gzdopen(dup_fd, "rb");
+	if (zfh == NULL) {
+	    throw ReadError("gzdopen() failed");
+	}
+	std::string out;
 	char buf[8192];
 	while (true) {
 	    int r = gzread(zfh, buf, sizeof(buf));
@@ -268,6 +266,20 @@ class DirectoryIterator {
 	gzclose(zfh);
 	return out;
     }
+
+    int get_fd() {
+	if (fd < 0) {
+	    open_fd();
+	} else {
+	    if (lseek(fd, 0, SEEK_SET) < 0)
+		throw CommitAndExit("Can't rewind file descriptor", path, errno);
+	}
+	return fd;
+    }
+
+    bool md5(std::string& out) {
+	return md5_fd(get_fd(), out);
+    }
 };
 
 #endif // OMEGA_INCLUDED_DIRITOR_H
