@@ -3,7 +3,7 @@
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2001 James Aylett
  * Copyright 2001,2002 Ananova Ltd
- * Copyright 2002,2003,2004,2006,2007,2008,2009,2010,2011,2014,2015,2016 Olly Betts
+ * Copyright 2002,2003,2004,2006,2007,2008,2009,2010,2011,2014,2015,2016,2018 Olly Betts
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -64,9 +64,6 @@ Xapian::RSet rset;
 
 map<string, string> option;
 
-string date_start, date_end, date_span;
-Xapian::valueno date_value_slot = Xapian::BAD_VALUENO;
-
 bool set_content_type = false;
 
 bool suppress_http_headers = false;
@@ -97,13 +94,38 @@ map_dbname_to_dir(const string &database_name)
     return database_dir + database_name;
 }
 
+// Get database(s) to search.
+template<typename IT>
+void
+parse_db_params(const pair<IT, IT>& dbs)
+{
+    dbname.resize(0);
+    // Only add a repeated db once.
+    set<string> seen;
+    for (auto i = dbs.first; i != dbs.second; ++i) {
+	const string& v = i->second;
+	if (v.empty()) continue;
+	size_t p = 0, q;
+	while (true) {
+	    q = v.find('/', p);
+	    string s(v, p, q - p);
+	    if (!s.empty() && seen.find(s) == seen.end()) {
+		// Translate DB parameter to path of database directory
+		if (!dbname.empty()) dbname += '/';
+		dbname += s;
+		db.add_database(Xapian::Database(map_dbname_to_dir(s)));
+		seen.insert(s);
+	    }
+	    if (q == string::npos) break;
+	    p = q + 1;
+	}
+    }
+}
+
 int main(int argc, char *argv[])
 try {
     read_config_file();
 
-    MCI val;
-    pair<MCI, MCI> g;
-
     option["flag_default"] = "true";
 
     // set default thousands and decimal separators: e.g. "16,729 hits" "1.4K"
@@ -139,29 +161,7 @@ try {
     }
 
     try {
-	// get database(s) to search
-	dbname.resize(0);
-	set<string> seen; // only add a repeated db once
-	g = cgi_params.equal_range("DB");
-	for (MCI i = g.first; i != g.second; ++i) {
-	    const string & v = i->second;
-	    if (!v.empty()) {
-		size_t p = 0, q;
-		while (true) {
-		    q = v.find('/', p);
-		    string s(v, p, q - p);
-		    if (!s.empty() && seen.find(s) == seen.end()) {
-			// Translate DB parameter to path of database directory
-			if (!dbname.empty()) dbname += '/';
-			dbname += s;
-			db.add_database(Xapian::Database(map_dbname_to_dir(s)));
-			seen.insert(s);
-		    }
-		    if (q == string::npos) break;
-		    p = q + 1;
-		}
-	    }
-	}
+	parse_db_params(cgi_params.equal_range("DB"));
 	if (dbname.empty()) {
 	    dbname = default_db;
 	    db.add_database(Xapian::Database(map_dbname_to_dir(dbname)));
@@ -172,7 +172,7 @@ try {
     }
 
     hits_per_page = 0;
-    val = cgi_params.find("HITSPERPAGE");
+    auto val = cgi_params.find("HITSPERPAGE");
     if (val != cgi_params.end()) hits_per_page = atol(val->second.c_str());
     if (hits_per_page == 0) {
 	hits_per_page = 10;
@@ -224,8 +224,8 @@ try {
 	// add expand/topterms terms if appropriate
 	string expand_terms;
 	if (cgi_params.find("ADD") != cgi_params.end()) {
-	    g = cgi_params.equal_range("X");
-	    for (MCI i = g.first; i != g.second; ++i) {
+	    auto g = cgi_params.equal_range("X");
+	    for (auto i = g.first; i != g.second; ++i) {
 		const string & v = i->second;
 		if (!v.empty()) {
 		    if (!expand_terms.empty())
@@ -236,8 +236,8 @@ try {
 	}
 
 	// collect the unprefixed prob fields
-	g = cgi_params.equal_range("P");
-	for (MCI i = g.first; i != g.second; ++i) {
+	auto g = cgi_params.equal_range("P");
+	for (auto i = g.first; i != g.second; ++i) {
 	    const string & v = i->second;
 	    if (!v.empty()) {
 		// If there are expand terms, append them to the first
@@ -259,9 +259,9 @@ try {
 	}
     }
 
-    g.first = cgi_params.lower_bound("P.");
-    g.second = cgi_params.lower_bound("P/"); // '/' is '.' + 1.
-    for (MCI i = g.first; i != g.second; ++i) {
+    auto begin = cgi_params.lower_bound("P.");
+    auto end = cgi_params.lower_bound("P/"); // '/' is '.' + 1.
+    for (auto i = begin; i != end; ++i) {
 	const string & v = i->second;
 	if (!v.empty()) {
 	    string pfx(i->first, 2, string::npos);
@@ -270,10 +270,10 @@ try {
     }
 
     // set any boolean filters
-    g = cgi_params.equal_range("B");
+    auto g = cgi_params.equal_range("B");
     if (g.first != g.second) {
 	vector<string> filter_v;
-	for (MCI i = g.first; i != g.second; ++i) {
+	for (auto i = g.first; i != g.second; ++i) {
 	    const string & v = i->second;
 	    // we'll definitely get empty B fields from "-ALL-" options
 	    if (!v.empty() && C_isalnum(v[0])) {
@@ -312,7 +312,7 @@ try {
     g = cgi_params.equal_range("N");
     if (g.first != g.second) {
 	vector<string> filter_v;
-	for (MCI i = g.first; i != g.second; ++i) {
+	for (auto i = g.first; i != g.second; ++i) {
 	    const string & v = i->second;
 	    // we'll definitely get empty N fields from "-ALL-" options
 	    if (!v.empty() && C_isalnum(v[0])) {
@@ -350,6 +350,60 @@ try {
     }
 
     // date range filters
+    struct date_range {
+	string start, end, span;
+    };
+    map<Xapian::valueno, date_range> date_ranges;
+    begin = cgi_params.lower_bound("START.");
+    end = cgi_params.lower_bound("START/"); // '/' is '.' + 1.
+    for (auto i = begin; i != end; ++i) {
+	const string & v = i->second;
+	if (!v.empty()) {
+	    Xapian::valueno slot = atoi(i->first.c_str() +
+					CONST_STRLEN("START."));
+	    date_ranges[slot].start = v;
+	}
+    }
+    begin = cgi_params.lower_bound("END.");
+    end = cgi_params.lower_bound("END/"); // '/' is '.' + 1.
+    for (auto i = begin; i != end; ++i) {
+	const string & v = i->second;
+	if (!v.empty()) {
+	    Xapian::valueno slot = atoi(i->first.c_str() +
+					CONST_STRLEN("END."));
+	    date_ranges[slot].end = v;
+	}
+    }
+    begin = cgi_params.lower_bound("SPAN.");
+    end = cgi_params.lower_bound("SPAN/"); // '/' is '.' + 1.
+    for (auto i = begin; i != end; ++i) {
+	const string & v = i->second;
+	if (!v.empty()) {
+	    Xapian::valueno slot = atoi(i->first.c_str() +
+					CONST_STRLEN("SPAN."));
+	    date_ranges[slot].span = v;
+	}
+    }
+    if (!date_ranges.empty()) {
+	// old_filters predates START.N, END.N and SPAN.N so use of any of
+	// these means this is definitely a different query.
+	old_filters.clear();
+    }
+    for (auto i : date_ranges) {
+	auto slot = i.first;
+	auto r = i.second;
+	add_date_filter(r.start, r.end, r.span, slot);
+	filters += '$';
+	filters += str(slot);
+	filters += '$';
+	filters += r.start;
+	filters += '$';
+	filters += r.end;
+	filters += '$';
+	filters += r.span;
+    }
+
+    string date_start, date_end, date_span;
     val = cgi_params.find("START");
     if (val != cgi_params.end()) date_start = val->second;
     val = cgi_params.find("END");
@@ -357,10 +411,12 @@ try {
     val = cgi_params.find("SPAN");
     if (val != cgi_params.end()) date_span = val->second;
     val = cgi_params.find("DATEVALUE");
+    Xapian::valueno date_value_slot = Xapian::BAD_VALUENO;
     if (val != cgi_params.end()) date_value_slot = string_to_int(val->second);
+    add_date_filter(date_start, date_end, date_span, date_value_slot);
 
     // If more default_op values are supported, encode them as non-alnums
-    // other than filter_sep or '!'.
+    // other than filter_sep, '!' or '$'.
     filters += (default_op == Xapian::Query::OP_AND ? '.' : '-');
     filters += date_start;
     filters += filter_sep;
