@@ -35,7 +35,7 @@
 
 #include <cassert>
 #include <cctype>
-#include "safeerrno.h"
+#include <cerrno>
 #include <stdio.h>
 #include <cstdlib>
 #include <cstring>
@@ -54,6 +54,7 @@
 #include "csvescape.h"
 #include "date.h"
 #include "datevalue.h"
+#include "fields.h"
 #include "jsonescape.h"
 #include "utils.h"
 #include "omega.h"
@@ -61,6 +62,7 @@
 #include "cgiparam.h"
 #include "loadfile.h"
 #include "sample.h"
+#include "sort.h"
 #include "str.h"
 #include "stringutils.h"
 #include "transform.h"
@@ -113,6 +115,12 @@ static Xapian::Query query;
 //static string url_query_string;
 Xapian::Query::op default_op = Xapian::Query::OP_AND; // default matching mode
 
+// Maintain an explicit date_filter_set flag - date_filter.empty() will also
+// be true if a date filter is specified which simplies to Query::MatchNothing
+// at construction time.
+static bool date_filter_set = false;
+static Xapian::Query date_filter;
+
 static Xapian::QueryParser qp;
 static Xapian::NumberRangeProcessor * size_rp = NULL;
 static Xapian::Stem *stemmer = NULL;
@@ -325,7 +333,25 @@ static querytype
 parse_queries(const string& oldp)
 {
     // Parse the query string.
-    qp.set_stemming_strategy(option["stem_all"] == "true" ? Xapian::QueryParser::STEM_ALL : Xapian::QueryParser::STEM_SOME);
+    auto opt_it = option.find("stem_strategy");
+    if (opt_it != option.end()) {
+	if (opt_it->second == "all") {
+	    qp.set_stemming_strategy(Xapian::QueryParser::STEM_ALL);
+	} else if (opt_it->second == "all_z") {
+	    qp.set_stemming_strategy(Xapian::QueryParser::STEM_ALL_Z);
+	} else if (opt_it->second == "none") {
+	    qp.set_stemming_strategy(Xapian::QueryParser::STEM_NONE);
+	} else if (opt_it->second == "some") {
+	    qp.set_stemming_strategy(Xapian::QueryParser::STEM_SOME);
+	} else if (opt_it->second == "some_full_pos") {
+	    qp.set_stemming_strategy(Xapian::QueryParser::STEM_SOME_FULL_POS);
+	}
+    } else {
+	opt_it = option.find("stem_all");
+	if (opt_it != option.end() && opt_it->second == "true") {
+	    qp.set_stemming_strategy(Xapian::QueryParser::STEM_ALL);
+	}
+    }
     qp.set_stopper(new MyStopper());
     qp.set_default_op(default_op);
     qp.set_database(db);
@@ -439,8 +465,6 @@ parse_queries(const string& oldp)
 static multimap<string, string> filter_map;
 static set<string> neg_filters;
 
-typedef multimap<string, string>::const_iterator FMCI;
-
 void add_bterm(const string &term) {
     string prefix;
     if (prefix_from_term(&prefix, term) > 0)
@@ -452,6 +476,39 @@ void add_nterm(const string &term) {
 	neg_filters.insert(term);
 }
 
+void
+add_date_filter(const string& date_start,
+		const string& date_end,
+		const string& date_span,
+		Xapian::valueno date_value_slot)
+{
+    if (date_start.empty() && date_end.empty() && date_span.empty())
+	return;
+
+    Xapian::Query q;
+    if (date_value_slot != Xapian::BAD_VALUENO) {
+	// The values can be a time_t in 4 bytes, or YYYYMMDD... (with the
+	// latter the sort order just works correctly between different
+	// precisions).
+	bool as_time_t =
+	    db.get_value_lower_bound(date_value_slot).size() == 4 &&
+	    db.get_value_upper_bound(date_value_slot).size() == 4;
+	q = date_value_range(as_time_t, date_value_slot,
+			     date_start, date_end,
+			     date_span);
+    } else {
+	q = date_range_filter(date_start, date_end, date_span);
+	q |= Xapian::Query("Dlatest");
+    }
+
+    if (date_filter_set) {
+	date_filter &= q;
+    } else {
+	date_filter_set = true;
+	date_filter = q;
+    }
+}
+
 static void
 run_query()
 {
@@ -463,7 +520,7 @@ run_query()
 	vector<Xapian::Query> filter_vec;
 	vector<string> same_vec;
 	string current;
-	for (FMCI i = filter_map.begin(); ; ++i) {
+	for (auto i = filter_map.begin(); ; ++i) {
 	    bool over = (i == filter_map.end());
 	    if (over || i->first != current) {
 		switch (same_vec.size()) {
@@ -510,25 +567,7 @@ run_query()
 	}
     }
 
-    if (!date_start.empty() || !date_end.empty() || !date_span.empty()) {
-	Xapian::Query date_filter;
-	if (date_value_slot != Xapian::BAD_VALUENO) {
-	    // The values can be a time_t in 4 bytes, or YYYYMMDD... (with the
-	    // latter the sort order just works correctly between different
-	    // precisions).
-	    bool as_time_t =
-		db.get_value_lower_bound(date_value_slot).size() == 4 &&
-		db.get_value_upper_bound(date_value_slot).size() == 4;
-	    date_filter = date_value_range(as_time_t, date_value_slot,
-					   date_start, date_end,
-					   date_span);
-	} else {
-	    date_filter = date_range_filter(date_start, date_end, date_span);
-	    date_filter = Xapian::Query(Xapian::Query::OP_OR,
-					date_filter,
-					Xapian::Query("Dlatest"));
-	}
-
+    if (date_filter_set) {
 	// If no query strings were provided then promote the daterange
 	// filter to be THE query instead of filtering an empty query.
 	if (query.empty()) {
@@ -544,7 +583,7 @@ run_query()
 	Xapian::Query filter(Xapian::Query::OP_OR,
 			     neg_filters.begin(), neg_filters.end());
 
-	if (query.empty()) {
+	if (query.empty() && !date_filter_set) {
 	    // If we only have a negative filter for the query, use MatchAll as
 	    // the query to apply the filters to.
 	    query = Xapian::Query::MatchAll;
@@ -692,14 +731,14 @@ string WordList::prev_list;
 unordered_map<string, int> WordList::word_to_occurrence;
 
 // Not a character in an identifier
-inline static bool
+static inline bool
 p_notid(unsigned int c)
 {
     return !C_isalnum(c) && c != '_';
 }
 
 // Not a character in an HTML tag name
-inline static bool
+static inline bool
 p_nottag(unsigned int c)
 {
     return !C_isalnum(c) && c != '.' && c != '-';
@@ -842,62 +881,24 @@ print_query_string(const char *after)
 }
 #endif
 
-class Fields {
-    mutable Xapian::docid did_cached;
-    mutable map<string, string> fields;
-
-    void read_fields(Xapian::docid did) const;
+class CachedFields : private Fields {
+    Xapian::docid did_cached = 0;
 
   public:
-    Fields() : did_cached(0) { }
+    CachedFields() {}
 
-    const string & get_field(Xapian::docid did, const string & field) const {
-	if (did != did_cached) read_fields(did);
-	return fields[field];
+    const string& get_field(Xapian::docid did, const string& name) {
+	if (did != did_cached) {
+	    did_cached = did;
+	    auto it = option.find("fieldnames");
+	    Fields::parse_fields(db.get_document(did).get_data(),
+				 it == option.end() ? nullptr : &it->second);
+	}
+	return Fields::get_field(name);
     }
 };
 
-void
-Fields::read_fields(Xapian::docid did) const
-{
-    fields.clear();
-    did_cached = did;
-    const string & data = db.get_document(did).get_data();
-
-    // Parse document data.
-    string::size_type i = 0;
-    const string & names = option["fieldnames"];
-    if (!names.empty()) {
-	// Each line is a field, with fieldnames taken from corresponding
-	// entries in the tab-separated list specified by $opt{fieldnames}.
-	string::size_type n = 0;
-	do {
-	    string::size_type n0 = n;
-	    n = names.find('\t', n);
-	    string::size_type i0 = i;
-	    i = data.find('\n', i);
-	    fields.insert(make_pair(names.substr(n0, n - n0),
-				    data.substr(i0, i - i0)));
-	} while (++n && ++i);
-    } else {
-	// Each line is a field, in the format NAME=VALUE.  We assume the field
-	// name doesn't contain an "=".  Lines without an "=" are currently
-	// just ignored.
-	do {
-	    string::size_type i0 = i;
-	    i = data.find('\n', i);
-	    string line(data, i0, i - i0);
-	    string::size_type j = line.find('=');
-	    if (j != string::npos) {
-		string & value = fields[line.substr(0, j)];
-		if (!value.empty()) value += '\t';
-		value.append(line, j + 1, string::npos);
-	    }
-	} while (++i);
-    }
-}
-
-static Fields fields;
+static CachedFields fields;
 static Xapian::docid q0;
 static Xapian::doccount hit_no;
 static int percent;
@@ -926,9 +927,9 @@ CMD_dbsize,
 CMD_def,
 CMD_defaultop,
 CMD_div,
-CMD_eq,
 CMD_emptydocs,
 CMD_env,
+CMD_eq,
 CMD_error,
 CMD_field,
 CMD_filesize,
@@ -1068,8 +1069,8 @@ T(defaultop,	   0, 0, N, 0), // default operator: "and" or "or"
 T(div,		   2, 2, N, 0), // integer divide
 T(emptydocs,	   0, 1, N, 0), // list of empty documents
 T(env,		   1, 1, N, 0), // environment variable
-T(error,	   0, 0, N, 0), // error message
 T(eq,		   2, 2, N, 0), // test equality
+T(error,	   0, 0, N, 0), // error message
 T(field,	   1, 2, N, 0), // lookup field in record
 T(filesize,	   1, 1, N, 0), // pretty printed filesize
 T(filters,	   0, 0, N, 0), // serialisation of current filters
@@ -1360,14 +1361,13 @@ eval(const string &fmt, const vector<string> &param)
 		break;
 	    }
 	    case CMD_cgi: {
-		MCI i = cgi_params.find(args[0]);
+		auto i = cgi_params.find(args[0]);
 		if (i != cgi_params.end()) value = i->second;
 		break;
 	    }
 	    case CMD_cgilist: {
-		pair<MCI, MCI> g;
-		g = cgi_params.equal_range(args[0]);
-		for (MCI i = g.first; i != g.second; ++i) {
+		auto g = cgi_params.equal_range(args[0]);
+		for (auto i = g.first; i != g.second; ++i) {
 		    value += i->second;
 		    value += '\t';
 		}
@@ -1473,9 +1473,6 @@ eval(const string &fmt, const vector<string> &param)
 		}
 		break;
 	    }
-	    case CMD_eq:
-		if (args[0] == args[1]) value = "true";
-		break;
 	    case CMD_emptydocs: {
 		string t;
 		if (!args.empty())
@@ -1493,6 +1490,9 @@ eval(const string &fmt, const vector<string> &param)
 		if (env != NULL) value = env;
 		break;
 	    }
+	    case CMD_eq:
+		if (args[0] == args[1]) value = "true";
+		break;
 	    case CMD_error:
 		if (error_msg.empty() && enquire == NULL && !dbname.empty()) {
 		    error_msg = "Database '" + dbname + "' couldn't be opened";
@@ -1672,7 +1672,7 @@ eval(const string &fmt, const vector<string> &param)
 		    }
 		}
 		// add any boolean terms
-		for (FMCI i = filter_map.begin(); i != filter_map.end(); ++i) {
+		for (auto i = filter_map.begin(); i != filter_map.end(); ++i) {
 		    url_query_string += "&B=";
 		    url_query_string += i->second;
 		}
@@ -2034,7 +2034,7 @@ eval(const string &fmt, const vector<string> &param)
 		// document id if relevant; empty otherwise
 		Xapian::docid id = q0;
 		if (!args.empty()) id = string_to_int(args[0]);
-		map<Xapian::docid, bool>::iterator i = ticked.find(id);
+		auto i = ticked.find(id);
 		if (i != ticked.end()) {
 		    i->second = false; // icky side-effect
 		    value = str(id);
@@ -2042,10 +2042,9 @@ eval(const string &fmt, const vector<string> &param)
 		break;
 	    }
 	    case CMD_relevants:	{
-		for (map <Xapian::docid, bool>::const_iterator i = ticked.begin();
-		     i != ticked.end(); ++i) {
-		    if (i->second) {
-			value += str(i->first);
+		for (auto i : ticked) {
+		    if (i.second) {
+			value += str(i.first);
 			value += '\t';
 		    }
 		}
@@ -2123,63 +2122,9 @@ eval(const string &fmt, const vector<string> &param)
 				     "<strong>", "</strong>", "...");
 		break;
 	    }
-	    case CMD_sort: {
-		const string &list = args[0];
-		if (list.empty()) break;
-		bool uniq = false;
-		bool rev = false;
-		if (args.size() > 1) {
-		    for (auto opt_ch : args[1]) {
-			switch (opt_ch) {
-			    case 'r':
-				rev = true;
-				break;
-			    case 'u':
-				uniq = true;
-				break;
-			    default:
-				throw string("Unknown $sort option: ") + opt_ch;
-			}
-		    }
-		}
-		vector<string> items;
-		string::size_type split = 0, split2;
-		do {
-		    split2 = list.find('\t', split);
-		    items.emplace_back(list, split, split2 - split);
-		    split = split2 + 1;
-		} while (split2 != string::npos);
-
-		if (!rev) {
-		    sort(items.begin(), items.end());
-		} else {
-		    sort(items.begin(), items.end(),
-			 [](const string& a, const string& b) {
-			     return a > b;
-			 });
-		}
-
-		value.reserve(list.size());
-		bool tab = false;
-		const string* prev = nullptr;
-		for (auto&& item : items) {
-		    // Skip duplicates if "u" flag specified.
-		    if (prev && *prev == item) {
-			continue;
-		    }
-		    if (uniq) {
-			prev = &item;
-		    }
-
-		    if (tab) {
-			value += '\t';
-		    } else {
-			tab = true;
-		    }
-		    value += item;
-		}
+	    case CMD_sort:
+		omegascript_sort(args, value);
 		break;
-	    }
 	    case CMD_split: {
 		string split;
 		if (args.size() == 1) {
@@ -2597,9 +2542,6 @@ ensure_query_parsed()
     if (query_parsed) return;
     query_parsed = true;
 
-    MCI val;
-    pair<MCI, MCI> g;
-
     // Should we discard the existing R-set recorded in R CGI parameters?
     bool discard_rset = false;
 
@@ -2609,7 +2551,7 @@ ensure_query_parsed()
 
     string v;
     // get list of terms from previous iteration of query
-    val = cgi_params.find("xP");
+    auto val = cgi_params.find("xP");
     if (val != cgi_params.end()) {
 	v = val->second;
 	// If xP given, default to discarding any RSet and forcing the first
@@ -2683,8 +2625,8 @@ ensure_query_parsed()
 
     if (!discard_rset) {
 	// put documents marked as relevant into the rset
-	g = cgi_params.equal_range("R");
-	for (MCI i = g.first; i != g.second; ++i) {
+	auto g = cgi_params.equal_range("R");
+	for (auto i = g.first; i != g.second; ++i) {
 	    const string & value = i->second;
 	    for (size_t j = 0; j < value.size(); j = value.find('.', j)) {
 		while (value[j] == '.') ++j;
